<?php
/**
 * @file
 * Provide node functionality.
 *
 * Copyright 2008-2009 by Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/**
 * Implementation of hook_node_info().
 */
function multivariate_node_info() {
  return array(
    'multivariate' => array(
      'name' => t('Multivariate study'),
      'module' => 'multivariate',
      'description' => t('Multivariate study.'),
      'body_label' => t('Description'),
      'custom' => TRUE,
      'locked' => TRUE
    ),
  );
}

/**
 * Implementation of hook_access().
 */
function multivariate_access($op, $node, $account) {
  $type = is_object($node) ? $node->type : $node;
  switch ($op) {
    case 'create':
      return user_access("create $type content", $account);
    case 'update':
      return user_access("edit own $type content", $account) || user_access("edit any $type content", $account);
    case 'delete':
      return user_access("delete own $type content", $account) || user_access("delete any $type content", $account);
    case 'view':
      if (user_access('access all multivariate content')) {
        return TRUE;
      }
      else {
        return user_access('access content', $account);
      }
  }
}

/**
 * Implementation of hook_form().
 */
function multivariate_form(&$node, $form_state) {
  drupal_add_css(drupal_get_path('module', 'multivariate') . '/multivariate.css');
  drupal_add_js(drupal_get_path('module', 'multivariate') . '/multivariate.js');

  $form = array();

  $type = node_get_types('type', $node);

  $form['basic'] = array(
    '#type' => 'fieldset',
    '#title' => t('Basic information'),
    '#weight' => -20,
  );
  $form['basic']['title'] = array(
    '#type' => 'textfield',
    '#title' => check_plain($type->title_label),
    '#required' => TRUE,
    '#default_value' => $node->title,
    '#weight' => -10
  );
  $form['basic']['body_field'] = node_body_field($node, $type->body_label, $type->min_word_count);
  $form['basic']['body_field']['body']['#rows'] = 5;

  $options = array();
  if (user_access('use multivariate testing')) {
    $options[] = MULTIVARIATE_TYPE_MUTATION;
  }
  if (user_access('use ab testing')) {
    $options[] = MULTIVARIATE_TYPE_VARIATION;
  }

  $form['basic']['study_type'] = array(
    '#type' => 'radios',
    '#title' => t('Type'),
    '#description' => t('The type of multivariate study to conduct. <b>Please note that if the type is changed any recorded data will be lost.</b>'),
    '#options' => $options,
    '#default_value' => isset($node->study_type) ? $node->study_type : '',
    '#required' => TRUE,
  );
  $form['basic']['study_type']['#options'] = drupal_map_assoc($form['basic']['study_type']['#options'], 'multivariate_type');

  $form['basic']['enabled'] = array(
    '#type' => 'checkbox',
    '#title' => t('Enabled'),
    '#default_value' => isset($node->enabled) ? $node->enabled : '',
  );

  multivariate_form_items('mutation', $node, $form_state, $form);

  multivariate_form_items('variation', $node, $form_state, $form);

  $form['success'] = array(
    '#type' => 'fieldset',
    '#title' => t('Success'),
    '#description' => t('The conditions that define a success. If more complex conditions are required please use ... hook.'),
    '#attributes' => array('id' => 'multivariate-success'),
    '#weight' => -8,
  );
  $form['success']['success_url'] = array(
    '#type' => 'textfield',
    '#title' => t('URL'),
    '#description' => t('The relative, Drupal URL that will be conserded a success if the user visits the page. Usually the page visited after a purchase or the like.'),
    '#default_value' => isset($node->success_url) ? $node->success_url : '',
  );

  return $form;
}

/**
 * Add the items for the specific type to the form.
 *
 * @param string $type Type of items, either 'mutation' or 'variation'.
 * @param object $node Node object to create items for.
 */
function multivariate_form_items($type, &$node, &$form_state, &$form) {
  $form[$type] = array(
    '#type' => 'fieldset',
    '#title' => t(ucfirst($type)),
    '#description' => t('<b>Please note that if any changes are made any recorded data will be lost.</b>'),
    '#weight' => -19,
    '#tree' => TRUE,
    '#prefix' => '<div id="multivariate-' . $type . '">',
    '#suffix' => '</div>',
  );
  if ($type == 'mutation') {
    $form[$type]['#description'] = t('Each mutation represents a change and the number associated with it is the number of states the change can have.') .
      ' ' . $form[$type]['#description'];
  }
  else {
    $form[$type]['#description'] = t('At the original URL a mutation will be selected which will redirect the user to one of the variation or leave them on the original page.') .
      ' ' . $form[$type]['#description'];
  }

  // Start count of new items at 1, and incrament on "Add $type" press.
  $count = isset($form_state['values'][$type]['count']) ? $form_state['values'][$type]['count'] : 1;
  if (isset($form_state['values'][$type]['add'])) {
    $count++;
  }
  $form[$type]['count'] = array(
    '#type' => 'value',
    '#value' => $count,
  );

  if (!isset($node->$type)) {
    $node->$type = array();
  }

  // Keep track of non-saved items.
  $items = &$node->$type;
  $offset = count($node->$type);
  for ($i = 1; $i <= $count; $i++) {
    $key = $i * -1;
    $value = isset($form_state['values'][$type][$key]) ? $form_state['values'][$type][$key] : '';
    $items[$key] = $value;
  }

  // Generation form elements for each item.
  $i = 0;
  foreach ($node->$type as $key => $value) {
    $form[$type][$key] = array(
      '#default_value' => $value,
      '#attributes' => array('class' => 'container-inline'),
    );
    if ($type == 'mutation') {
      $form[$type][$key] += array(
        '#type' => 'select',
        '#title' => t('Mutation @id', array('@id' => $i)),
        '#options' => array(0 => '--') + drupal_map_assoc(array(2, 3, 4, 5, 6)),
      );
    }
    elseif ($type == 'variation') {
      $form[$type][$key] += array(
        '#type' => 'textfield',
        '#title' => $i == 0 ? t('Original') : t('Variation @id', array('@id' => $i)),
      );
    }
    $i++;
  }
  $form[$type]['add'] = array(
    '#type' => 'submit',
    '#value' => t('Add @type', array('@type' => $type)),
    '#ahah' => array(
      'path' => 'multivariate/add/' . $type,
      'wrapper' => 'multivariate-' . $type,
      'method' => 'replace',
      'effect' => 'fade',
    ),
    '#submit' => array('multivariate_add_submit'),
  );
}

/**
 * Add variation AHAH callback.
 *
 * @link http://drupal.org/node/331941
 */
function multivariate_add_ahah($type) {
  include_once 'modules/node/node.pages.inc';

  $form_state = array('storage' => NULL, 'submitted' => FALSE);
  $form_build_id = $_POST['form_build_id'];
  $form = form_get_cache($form_build_id, $form_state);

  $args = $form['#parameters'];
  $form_id = array_shift($args);
  $form_state['post'] = $form['#post'] = $_POST;
  $form['#programmed'] = $form['#redirect'] = FALSE;

  drupal_process_form($form_id, $form, $form_state);
  $form = drupal_rebuild_form($form_id, $form_state, $args, $form_build_id);

  unset($form[$type]['#prefix'], $form[$type]['#suffix']);

  // TODO Removes status messages for required fields.
  drupal_get_messages();
  drupal_json(array(
    'status' => TRUE,
    'data' => theme('status_messages') . drupal_render($form[$type])
  ));
}

/**
 * Add action submit handler.
 */
function multivariate_add_submit($form, &$form_state) {
  $form_state['rebuild'] = TRUE;
}

/**
 * Implementation of hook_load().
 */
function multivariate_load($node) {
  $additions = db_fetch_object(db_query('SELECT * FROM {multivariate} WHERE vid = %d', $node->vid));
  $additions->mutation = unserialize($additions->mutation);
  $additions->variation = unserialize($additions->variation);
  return $additions;
}

/**
 * Implementation of hook_insert().
 */
function multivariate_insert($node) {
  multivariate_node_clean($node);
  multivariate_node_calculate($node);
  db_query("INSERT INTO {multivariate} (nid, vid, study_type, enabled, mutation, variation, success_url)
            VALUES (%d, %d, %d, %d, '%s', '%s', '%s')", $node->nid, $node->vid, $node->study_type, $node->enabled,
            serialize($node->mutation), serialize($node->variation), $node->success_url);
}

/**
 * Implementation of hook_update().
 */
function multivariate_update($node) {
  if (isset($node->revision) && $node->revision) {
    multivariate_insert($node);
    return;
  }

  multivariate_node_clean($node);
  multivariate_node_calculate($node);
  db_query("UPDATE {multivariate}
            SET study_type = %d,
            enabled = %d,
            mutation = '%s',
            variation = '%s',
            success_url = '%s'
            WHERE vid = %d", $node->study_type, $node->enabled, serialize($node->mutation),
            serialize($node->variation), $node->success_url, $node->vid);
}

/**
 * Clean the items in preparation to be saved.
 *
 * Remove 'count' and 'add items and renumber all items to ensure continuous
 * numbering.
 *
 * @param object $node Node object to clean.
 */
function multivariate_node_clean(&$node) {
  foreach (array('mutation', 'variation') as $type) {
    $items = &$node->$type;
    unset($items['count']);
    unset($items['add']);

    $array = $items;
    $items = array();
    foreach ($array as $key => $value) {
      if ($value) {
        $items[] = $value;
      }
    }
  }
}

/**
 * Calculate the possible mutations and store them.
 *
 * @param object $node Node object to calculate mutations for.
 */
function multivariate_node_calculate($node) {
  $node_old = node_load($node->nid);
  if ($node->study_type != $node_old->study_type
      || $node->mutation != $node_old->mutation
      || $node->variation != $node_old->variation) {
    db_query('DELETE FROM {multivariate_mutation} WHERE nid = %d', $node->nid);

    $mutations = multivariate_mutation_definition_get($node);
    $binary = multivariate_node_calculate_binary($mutations, 0);
    foreach ($binary as $state) {
      $mutation_id = bindec($state);
      db_query('INSERT INTO {multivariate_mutation} (nid, mutation_id, count)
                VALUES (%d, %d, %d, %d)', $node->nid, $mutation_id, 0, 0);
    }
  }
}

/**
 * Calculate the binary representations of each mutation.
 *
 * @param array $mutations List of mutations and their state count.
 * @param integer $active The active mutation.
 * @return array Set of binary representations.
 */
function multivariate_node_calculate_binary($mutations, $active) {
  $binary = array();

  $state_count = $mutations[$active];
  $base = implode('', array_fill(0, $state_count, '0'));
  // Cycle through the state of the mutation.
  for ($state = 0; $state < $state_count; $state++) {
    $prefix = $base;
    $prefix[$state] = '1';

    if ($active < count($mutations) - 1) {
      $subs = multivariate_node_calculate_binary($mutations, $active + 1);
      foreach ($subs as $sub) {
        $binary[] = $prefix . $sub;
      }
    }
    else {
      $binary[] = $prefix;
    }
  }
  return $binary;
}

/**
 * Implementation of hook_delete().
 */
function multivariate_delete($node) {
  db_query('DELETE FROM {multivariate} WHERE nid = %d', $node->nid);
}

/**
 * Implementation of hook_view().
 */
function multivariate_view($node, $teaser = FALSE, $page = FALSE) {
  $node = node_prepare($node, $teaser);
  $node->content['body']['#weight'] = -10;

  $node->content['study_type'] = array(
    '#value' => t('<b>Type:</b> @type<br />', array('@type' => multivariate_type($node->study_type))),
    '#weight' => -9,
  );
  $node->content['enabled'] = array(
    '#value' => t('<b>Enabled:</b> @enabled<br />', array('@enabled' => $node->enabled ? t('True') : t('False'))),
    '#weight' => -8,
  );

  $definition = multivariate_mutation_definition_get($node);
  $results = db_query('SELECT mutation_id, count, success
                      FROM {multivariate_mutation}
                      WHERE nid = %d', $node->nid);
  $data = array();
  while ($result = db_fetch_array($results)) {
    $binary = multivariate_mutation_correct(decbin($result['mutation_id']), $definition);
    $mutation = multivariate_mutation_parse($definition, $binary);

    foreach ($mutation as $key => $state) {
      if (!isset($data[$key])) {
        $data[$key] = array();
      }
      if (!isset($data[$key][$state])) {
        $data[$key][$state] = array(0, 0);
      }
      $data[$key][$state][0] += $result['success'];
      $data[$key][$state][1] += $result['count'];
    }
  }

  $node->content['results'] = array(
    '#type' => 'fieldset',
    '#title' => t('@type results', array('@type' => multivariate_type($node->study_type))),
  );

  $weight = -7;
  foreach ($data as $key => $states) {
    $chart = array(
      '#chart_id' => 'multivariate_results_' . $key,
      '#title' => t('Group @key', array('@key' => $key)),
      '#type' => CHART_TYPE_BAR_V_GROUPED,
      '#size' => chart_size(count($states) * 40 + 60, 150),
      '#bar_size' => chart_bar_size(40, 5),
      '#data_colors' => array(),
      '#data' => array(),
      '#mixed_axis_labels' => array(CHART_AXIS_Y_LEFT => array(array(chart_mixed_axis_range_label(0, 100)))),
    );

    ksort($states);
    foreach ($states as $state => $values) {
      $chart['#data'][$state][] = $values[1] > 0 ? ($values[0] / $values[1]) * 100 : 0;
      $chart['#data_colors'][$state] = chart_unique_color('s' . $state);
    }

    $node->content['results'][] = array(
      '#value' => chart_render($chart),
      '#weight' => $weight++,
    );
  }

  return $node;
}

/**
 * Implementation of hook_link().
 */
function multivariate_link($type, $object, $teaser = FALSE) {
  $links = array();

  if ($type == 'node') {
    $links['multivariate_export'] = array(
      'title' => t('export as csv'),
      'href' => 'node/' . $object->nid . '/multivariate/export',
    );
  }

  return $links;
}

/**
 * Export multivariate data.
 *
 * @param object $node Multivariate node object.
 */
function multivariate_node_export($node) {
  header('Content-Disposition: attachment; filename="multivariate-' . $node->nid . '.csv"');
  header('Content-Type: text/plain');

  $definition = multivariate_mutation_definition_get($node);
  $results = db_query('SELECT mutation_id, count, success
                       FROM {multivariate_mutation}
                       WHERE nid = %d', $node->nid);

  echo '"' . implode('","', array('mutation', 'count', 'success')) . '"' . "\n";
  while ($result = db_fetch_array($results)) {
    $binary = multivariate_mutation_correct(decbin($result['mutation_id']), $definition);
    echo '"' . implode('","', array($binary, $result['count'], $result['success'])) . '"' . "\n";
  }
}
